from virt_s3.utils import get_custom_logger
from virt_s3.main import get_default_params, get_valid_file_paths, get_file, get_session_client, upload_data, SessionManager
import virt_s3.s3 as s3
import virt_s3.fs as fs

from io import BytesIO
from abc import ABC, abstractmethod
from typing import List, Union, Any, Callable
import json

logger = get_custom_logger()

try: 
    import pandas as pd
except ModuleNotFoundError as me: 
    logger.warning(f"{me} -> `$ poetry install -E dataframe` to use.")
    class pd:
        class DataFrame: 
            pass
try: 
    import pyarrow.parquet as pq
except ModuleNotFoundError as me: 
    logger.warning(f"{me} -> `$ poetry install -E dataframe` to use.")

#######################################################################
################ Extra Functions, Classes, Interfaces #################
#######################################################################


def read_parquet_file_df(
    fpath: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    engine: str = "auto",
    columns: List[str] = None,
    filters: List[tuple] = None,
    use_nullable_dtypes: bool = False,
    **download_kwargs
) -> pd.DataFrame:
    """Convenience function to read parquet file as pandas DataFrame.
    Attempts to first read full buffer at once, then falls back to reading in batches if error
    in reading full buffer at once.

    :param fpath: path to parquet file or key
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :param engine: {{'auto', 'pyarrow', 'fastparquet'}}, defaults to 'auto'
        - Parquet library to use. If 'auto', then the option
        ``io.parquet.engine`` is used. The default ``io.parquet.engine``
        behavior is to try 'pyarrow', falling back to 'fastparquet' if
        'pyarrow' is unavailable. When using the ``'pyarrow'`` engine and no storage options are provided
        and a filesystem is implemented by both ``pyarrow.fs`` and ``fsspec`` (e.g. "s3://"), 
        then the ``pyarrow.fs`` filesystem is attempted first. Use the filesystem keyword with an
        instantiated fsspec filesystem if you wish to use its implementation.

    :param columns: If not None, only these columns will be read from the file, defaults to None
    :param filters: To filter out data, defaults to None
        - Filter syntax: [[(column, op, val), ...],...]
        where op is [==, =, >, >=, <, <=, !=, in, not in]
        The innermost tuples are transposed into a set of filters applied
        through an `AND` operation. The outer list combines these sets of filters through an `OR`
        operation. A single list of tuples can also be used, meaning that no `OR`
        operation between set of filters is to be conducted. Using this argument will NOT result in 
        row-wise filtering of the final partitions unless ``engine="pyarrow"`` is also specified.  For other engines, filtering 
        is only performed at the partition level, that is, to prevent the loading of some row-groups and/or files.

    :param use_nullable_dtypes: boolean flag, defaults to False
        - If True, use dtypes that use ``pd.NA`` as missing value indicator for the resulting DataFrame. 
        (only applicable for the ``pyarrow`` engine). As new dtypes are added that support ``pd.NA`` in the future, the
        output with this option will change to use those dtypes.
        Note: this is an experimental option, and behaviour (e.g. additional
        support dtypes) may change without notice.
    
    :return: pd.DataFrame
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)
    # get file as dataframe
    buffer = get_file(fpath, params=params, client=client, bytes_io=True, **download_kwargs)
    try:
        return pd.read_parquet(
            buffer,
            engine=engine,
            columns=columns,
            filters=filters,
            use_nullable_dtypes=use_nullable_dtypes
        )
    except Exception as e:
        logger.debug(f"Failed to read full parquet file from buffer. Falling back to reading in batches. {e}")
        parquet_file = pq.ParquetFile(buffer)
        dfs = [batch.to_pandas() for batch in parquet_file.iter_batches()]
        return pd.concat(dfs, ignore_index=True, copy=False)   


def write_parquet_file_df(
    df: pd.DataFrame,
    save_fpath: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    engine: str = "auto",
    compression: str = None,
    index: bool = None,
    partition_cols: List[str] = None,
    **upload_kwargs
) -> str:
    """Convenience function to write pandas DataFrame to parquet file.

    :param df: pandas DataFrame to save
    :param save_fpath: parquet file save path or key (e.g. data.parquet.sz)
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :param engine: {{'auto', 'pyarrow', 'fastparquet'}}, defaults to 'auto'
        - Parquet library to use. If 'auto', then the option
        ``io.parquet.engine`` is used. The default ``io.parquet.engine``
        behavior is to try 'pyarrow', falling back to 'fastparquet' if
        'pyarrow' is unavailable. When using the ``'pyarrow'`` engine and no storage options are provided
        and a filesystem is implemented by both ``pyarrow.fs`` and ``fsspec`` (e.g. "s3://"), 
        then the ``pyarrow.fs`` filesystem is attempted first. Use the filesystem keyword with an
        instantiated fsspec filesystem if you wish to use its implementation.

    :param compression: compression algorithm to use, default 'snappy'
        - Name of the compression to use. Use None for no compression. 
        - Supported options: 'snappy', 'gzip', 'brotli', 'lz4', 'zstd'.

    :param index: _description_, defaults to None
        - If True, include the dataframe's index(es) in the file output. 
        - If False, they will not be written to the file. 
        - If None, similar to True the dataframe's index(es) will be saved. 
        However, instead of being saved as values, the RangeIndex will be stored as 
        a range in the metadata so it doesn't require much space and is faster. 
        Other indexes will be included as columns in the file output.
    
    :param partition_cols: column names by which to partition the dataset, defaults to None
        - Columns are partitioned in the order they are given
        - Must be None if path is not a string

    :return: saved key or file path
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)
    # write to buffer
    buffer = BytesIO()
    df.to_parquet(buffer, engine=engine, compression=compression, partition_cols=partition_cols, index=index)
    buffer.seek(0)
    # upload
    return upload_data(buffer, save_fpath, params=params, client=client, **upload_kwargs)


class FileValidator(ABC):
    """Base Abstract Interface for File Type Validators
   
    :param file_extensions: supported file extensions by this Validator, defaults to [".json", ".csv", ".parquet.sz"]
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to get_default_params
    """
    def __init__(
        self, 
        file_extensions: List[str] = [".json", ".csv", ".parquet.sz"],
        params: Union[s3.S3Params, fs.LocalFSParams] = get_default_params,
    ) -> None:
        
        if isinstance(params, Callable):
            params = params()
        self.file_extensions = file_extensions
        self.params = params

        self.fpaths: List[str] = []
        self.bad_fpaths: List[str] = []
    
    @abstractmethod
    def validate_file(
        self, 
        fpath: str, 
        *args, 
        params: Union[s3.S3Params, fs.LocalFSParams] = None, 
        client: Union[s3.Session, None] = None,
        **kwargs
    ) -> bool:
        """Method to implement based on specific child validator

        :param fpath: path to file or key
        :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to get_default_params()
        :param client: instance of either s3.boto3.Sesssion or None, defaults to None
        :return: True if valid file, False if invalid file
        """
        pass
    
    def run(
        self, 
        fpath_li: List[str] = [],
        root_path: str = None, 
        ignore: List[str] = [],
    ) -> dict:
        """Method to run File Validator

        :param fpath_li: list of files paths to run validation on, defaults to []
        :param root_path: root dir path or prefix to start looking for files within to validated, defaults to None
        :param ignore: list of dirs/prefixes/files to ignore, defaults to []
        :return: dictionary of { 'valid': List[file paths], 'invalid': List[file paths] }
        """
        ## Check for valid arguments passed in 
        if root_path is None and (fpath_li is None or (isinstance(fpath_li, list) and len(fpath_li) < 1)):
            raise Exception(f"Invalid arguments passed in to FileValidator.run(). Either `fpath_li` or `root_path` must be passed in and non-null.")

        # initialize valid and invalid file paths
        self.fpaths, self.bad_fpaths = [], []

        ## File Path retrieval
        # if fpath list passed in, use that to read and validate files
        fpaths = fpath_li
        if len(fpaths) < 1: 
            with SessionManager(params=self.params) as client:
                # else, get current list of file paths
                fpaths = get_valid_file_paths(
                    root_path, 
                    ignore=ignore,
                    params=self.params, 
                    client=client,
                    filter_extensions=self.file_extensions
                )
                logger.info(f"Retrieved {len(fpaths)} Files with one of the following extensions: {self.file_extensions}")

        ## Filtering Extensions
        logger.debug(f"Checking to make sure all files include extensions: {self.file_extensions} ...")
        remove = []
        for fpath in fpath_li:
            correct_extension = False
            for extension in self.file_extensions:
                if fpath.endswith(extension):
                    correct_extension = True
            if not correct_extension:
                # logger.debug(f"Removing file from list due to non-inclusion of extensions {self.file_extensions}: {fpath}")
                remove.append(fpath)
        fpaths = [f for f in fpaths if f not in remove]
        logger.debug(f"Correct Files to Validate: {fpaths[:10]}, ... +{len(fpaths)-10}")

        ## Data Validation
        logger.info(f'Running Data Validation for {self.file_extensions} files ...')
        with SessionManager(params=self.params) as client:
            for i, fpath in enumerate(fpaths):
                is_valid = self.validate_file(fpath, params=self.params, client=client)
                if not is_valid:
                    self.bad_fpaths.append(fpath)
                    continue
                self.fpaths.append(fpath)

            logger.debug(f"Valid Files: {len(self.fpaths)}")
            logger.debug(f"Invalid Files: {len(self.bad_fpaths)}")

        return {'valid': self.fpaths, 'invalid': self.bad_fpaths}


class JSONFileValidator(FileValidator):
    """Interface for JSON FIle Validation
   
    :param file_extensions: supported file extensions by this Validator, defaults to [".json", ".csv", ".parquet.sz"]
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to get_default_params
    """
    def __init__(
        self, 
        file_extensions: List[str] = [".json"],
        params: Union[s3.S3Params, fs.LocalFSParams] = get_default_params,
    ) -> None:
        
        super().__init__(file_extensions=file_extensions, params=params)
    
    def validate_file(self, fpath: str, client: Union[s3.Session, None] = None) -> bool:
        """Implemented method to validate JSON files

        :param fpath: path to file or key
        :param client: instance of either s3.boto3.Sesssion or None, defaults to None
        :return: True if valid file, False if invalid file
        """
        try:
            obj = get_file(fpath, bytes_io=False, params=self.params, client=client)
            data = json.loads(obj.decode('utf-8'))
            assert len(list(data.keys())) > 0
            return True
        except Exception as e:
            logger.debug(f"JSONFileValidatorException: {e}")
            return False


class CSVFileValidator(FileValidator):
    """Interface for CSV File Validation
   
    :param file_extensions: supported file extensions by this Validator, defaults to [".json", ".csv", ".parquet.sz"]
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to get_default_params
    """
    def __init__(
        self, 
        file_extensions: List[str] = [".csv"],
        params: Union[s3.S3Params, fs.LocalFSParams] = get_default_params,
    ) -> None:
        
        try: import pandas as pd
        except: logger.warning(f"'pandas' not installed.  Run `$ poetry install -E dataframe` to use CSVFileValidator, ParquetFileValidator.")
        
        super().__init__(file_extensions=file_extensions, params=params)
    
    def validate_file(self, fpath: str, client: Union[s3.Session, None] = None) -> bool:
        """Implemented method to validate CSV files

        :param fpath: path to file or key
        :param client: instance of either s3.boto3.Sesssion or None, defaults to None
        :return: True if valid file, False if invalid file
        """
        try:
            obj = get_file(fpath, bytes_io=True, params=self.params, client=client)
            df = pd.read_csv(obj)
            assert isinstance(df, pd.DataFrame)
            return True
        except Exception as e:
            logger.debug(f"CSFileVValidatorException: {e}")
            return False
        

class ParquetFileValidator(FileValidator):
    """Interface for Parquet File Validation
   
    :param file_extensions: supported file extensions by this Validator, defaults to [".json", ".csv", ".parquet.sz"]
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to get_default_params
    """
    def __init__(
        self, 
        file_extensions: List[str] = [".parquet.sz", ".parquet"],
        params: Union[s3.S3Params, fs.LocalFSParams] = get_default_params,
    ) -> None:
        try: import pandas as pd
        except: logger.warning(f"'pandas' not installed.  Run `$ poetry install -E dataframe` to use CSVFileValidator, ParquetFileValidator.")
        try: import pyarrow.parquet as pq
        except: logger.warning(f"'pyarrow' not installed. Run `$ poetry install -E dataframe` to use ParquetFileValidator.")
            
        super().__init__(file_extensions=file_extensions, params=params)

    def validate_file(self, fpath: str, client: Union[s3.Session, None] = None) -> bool:
        """Implemented method to validate Parquet files

        :param fpath: path to file or key
        :param client: instance of either s3.boto3.Sesssion or None, defaults to None
        :return: True if valid file, False if invalid file
        """
        try:
            df = read_parquet_file_df(fpath, params=self.params, client=client)
            assert isinstance(df, pd.DataFrame)
            return True
        except Exception as e:
            logger.debug(f"ParquetFileValidatorException: {e}")
            return False


from virt_s3.utils import get_custom_logger, ImageFormatType, archive_zip_as_buffer, KB, MB
import virt_s3.s3 as s3
import virt_s3.fs as fs

from io import BytesIO
from typing import List, Union, Any, Optional, Generator
import os

logger = get_custom_logger()


#######################################################
################## Main API Functions #################
#######################################################


class SessionManager:
    """General Session Context Manager for virt_s3 repo

    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :param keep_open: boolean flag to keep the session open or not outside of the context manager scope, defaults to False
    :raises ValueError: if the params either passed in or retreived is not of type `fs.LocalFSParams` or `s3.S3Params`

    Example
    --------
    >>> params = virt_s3.get_default_params()
    >>> with virt_s3.SessionManager(params=params) as session:
    ...     virt_s3.create_bucket('default-bucket', params=params, client=session)
    ...
    """
    def __init__(
        self, 
        params: Union[s3.S3Params, fs.LocalFSParams] = None,
        client: Union[s3.Session, None] = None,
        keep_open: bool = False,
        **session_client_kwargs
    ) -> None:
        # if no params passed in get default params
        if params is None:  params = get_default_params()
        # create session client if not provided
        if client is None: client = get_session_client(params)
    
        # handle local fs function
        if isinstance(params, fs.LocalFSParams) and client is None:
            self.manager = fs.LocalFSSessionManager(
                local_fs_params=params, 
                client=client, 
                keep_open=keep_open, 
                **session_client_kwargs
            )
        elif isinstance(params, s3.S3Params) or isinstance(client, s3.Session):
            self.manager = s3.S3SessionManager(
                s3_params=params,
                s3_client=client,
                keep_open=keep_open,
                **session_client_kwargs
            )
        else:
            raise ValueError(f"Params is not of instance `fs.LocalFSParams` or `s3.S3Params`")

    def __enter__(self):
        return self.manager.__enter__()
    
    def __exit__(self, exc_type, exc_value, traceback):
        return self.manager.__exit__(exc_type, exc_value, traceback)


def get_default_params(
    use_local: bool = False, 
    use_s3: bool = False,
) -> Union[s3.S3Params, fs.LocalFSParams]:
    """Function to get default parameters to use for all functions (default behavior is based off of env variables)

    :param use_local: overrides env variables and forces to use local fs params, defaults to False
    :param use_s3: overrieds env variables and forces to use s3 params, defaults to False
    :return: either instance of s3.S3Params or fs.LocalFSParams
    """
    # check overrides
    if use_local: return fs.get_local_fs_params()
    elif use_s3: return s3.get_s3_default_params()

    # check env variables and return default beahvior
    local_fs_params = fs.get_local_fs_params()
    if local_fs_params.use_local_fs:
        if not os.path.exists(local_fs_params.root_dir):
            logger.error(f"Root Directory for Local File System Parameters Does not Exist!: {local_fs_params.root_dir}") 
        return local_fs_params
    
    return s3.get_s3_default_params()


def get_session_client(
    params: Union[s3.S3Params, fs.LocalFSParams] = None, 
    **kwargs
) -> Union[s3.Session, None]:
    """Function to get session client based on passed in s3.S3Params or fs.LocalFSParams

    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :return: instance of either s3.boto3.Sesssion or None
    """
    # local file system does not have concept of 'session client' so return None
    if params is None: params = get_default_params()
    if isinstance(params, fs.LocalFSParams): return
    elif isinstance(params, s3.S3Params): return s3.get_s3_session_client(params, **kwargs)
    return 


def create_bucket(
    bucket_name: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    **kwargs
) -> bool:
    """Function to create a bucket to read and write from

    :param bucket_name: name of bucket to create
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :return: True if successfully created (or already exists), else False
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.create_fs_bucket(bucket_name, local_fs_params=params)
    
    # handle s3 function
    return s3.create_s3_bucket(bucket_name, s3_params=params, s3_client=client, **kwargs)


def get_file_chunked(
    fpath: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    **kwargs
) -> Union[BytesIO, bytes, None]:
    """Function to get a file using a chunking loop. This can be useful when trying to 
    retrieve very large files

    :param fpath: file path or key path of object to get
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :return: either instance of BytesIO object, bytes object, or None
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.get_fs_file_chunked(fpath, local_fs_params=params, **kwargs)

    # handle s3 function
    return s3.get_s3_file_chunked(fpath, s3_params=params, s3_client=client, **kwargs)


def get_file(
    fpath: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    bytes_io: bool = False,
    **kwargs
) -> Union[BytesIO, bytes, None]:
    """Function to retrieve specified file

    :param fpath: file path or key path of object to get
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: either BytesIO object, bytes object, or None
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.get_fs_file(fpath, local_fs_params=params, bytes_io=bytes_io, **kwargs)
    
    # handle s3 function
    return s3.get_s3_file(fpath, s3_params=params, s3_client=client, bytes_io=bytes_io, **kwargs)


def get_image(
    fpath: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    img_format: Optional[ImageFormatType] = None,
    bytes_io: bool = False,
    **kwargs
) -> Union[BytesIO, s3.Image.Image, s3.np.ndarray, str, bytes, None]:
    """Function to get image from either s3 or local file system

    :param fpath: file path or key path of object to get
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :param img_format: one of ImageFormatType enum, defaults to None, defaults to None
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: an instance of either io.BytesIO, PIL.Image.Image, np.ndarray, string, bytes, or None
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.get_fs_image(fpath, img_format, local_fs_params=params, bytes_io=bytes_io, **kwargs)

    # handle s3 function
    return s3.get_s3_image(fpath, s3_params=params, s3_client=client, img_format=img_format, bytes_io=bytes_io, **kwargs)


def get_files_generator(
    fpath_li: List[str],
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    bytes_io: bool = False,
    **kwargs  
) -> Generator[Union[BytesIO, bytes, None], Any, None]:
    """Generator function to quickly loop through reading a list of keys or file paths

    :param fpath_li: list of key or file paths to retreive as generator loop
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: None
    :yield: either BytesIO object, bytes object, or None
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.get_fs_files_generator(fpath_li, local_fs_params=params, bytes_io=bytes_io, **kwargs)
    
    # handle s3 function
    return s3.get_s3_files_generator(fpath_li, s3_params=params, s3_client=client, bytes_io=bytes_io, **kwargs)


def get_files_batch(
    fpath_li: List[str],
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    max_concurrency: int = 0, 
    bytes_io: bool = False,
    **kwargs  
) -> List[Union[BytesIO, bytes, None]]:
    """Function to get list of file paths or key paths in batch

    :param fpath_li: list of key or file paths to retreive as generator loop
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :param max_concurrency: max no. of concurrent threads to use when downloading a file (0 means no concurrency), defaults to 0
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: list of either BytesIO object, bytes object, or None
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.get_fs_files_batch(fpath_li, local_fs_params=params, max_concurrency=max_concurrency, bytes_io=bytes_io, **kwargs)
    
    # handle s3 function
    return s3.get_s3_files_batch(fpath_li, s3_params=params, s3_client=client, max_concurrency=max_concurrency, bytes_io=bytes_io, **kwargs)


def list_dirs(
    dir_path: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    delimiter: str ='/',
    **kwargs
) -> List[str]:
    """Function to list valid folders within 'bucket'

    :param dir_path: directory or prefix path
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :param delimiter: delimiter used to filter objects in S3 (e.g. only subfolders - delimiter="/"), defaults to '/'
    :return: list of folder paths
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.list_fs_folders(dir_path, local_fs_params=params, **kwargs)
    
    # handle s3 functions
    return s3.list_s3_folders(dir_path, s3_params=params, s3_client=client, delimiter=delimiter, **kwargs)


def get_valid_file_paths(
    dir_path: str,
    ignore: List[str] = [".DS_Store"],
    filter_extensions: List[str] = [],
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    **kwargs
) -> List[str]:
    """Function to get list of valid file paths or keys within particular directory of bucket

    :param dir_path: directory or prefix path to search
    :param ignore: list of directories/prefixes and filepaths to ignore, defaults to ['.DS_Store']
    :param filter_extensions: list of file extensions to filter for (empty list means get everything), defaults to []
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :return: list of valid file paths or key paths
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.get_valid_fs_fpaths(
            dir_path, 
            local_fs_params=params, 
            ignore=ignore, 
            filter_extensions=filter_extensions,
            **kwargs
        )

    # handle s3 function
    return s3.get_valid_s3_keys(
        dir_path, 
        ignore=ignore, 
        filter_extensions=filter_extensions, 
        s3_params=params, 
        s3_client=client,
        **kwargs
    )


def file_exists(
    fpath: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    **kwargs
) -> bool:
    """Function to see if key or file path exists in bucket

    :param fpath: file path or key path to check
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :return: True if key or file path exists, False otherwise
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.fpath_exists_fs(fpath, local_fs_params=params, **kwargs)

    # handle s3 function
    return s3.key_exists_s3(fpath, s3_params=params, s3_client=client, **kwargs)


def upload_data(
    data: Union[bytes, memoryview, BytesIO],
    fpath: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    **kwargs
) -> str:
    """Function to upload data to S3 or local file system

    :param data: data to save as bytes, memoryview, or BytesIO
    :param fpath: path to save data to
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :raises Exception: either General exception if not HTTP Status Code 200 or Exception if input data is not of instance 'bytes' or 'BytesIO'
    :return: saved key or file path
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.upload_to_fs(data, fpath, local_fs_params=params, **kwargs)
    
    # handle s3 function
    return s3.upload_to_s3(data, fpath, s3_params=params, s3_client=client, **kwargs)


def delete_file(
    fpath: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    **kwargs
) -> bool:
    """Function to delete a file from s3 or local file system

    :param fpath: file path to delete
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :return: True if delete successful, else return False
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.delete_fs_file(fpath, local_fs_params=params, **kwargs)
    
    # handle s3 function
    return s3.delete_s3_object(fpath, s3_params=params, s3_client=client, **kwargs)


def delete_files_by_dir(
    dir_path: str,
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    **kwargs
) -> List[str]:
    """Function to delete all files and subdirectories, etc. in a given folder

    :param dir_path: path to folder or prefix to delete
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :return: list of deleted file or key paths
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.delete_fs_files_by_dir(dir_path, local_fs_params=params, **kwargs)
    
    # handle s3 function
    return s3.delete_s3_objects_by_prefix(dir_path, s3_params=params, s3_client=client, **kwargs)


def extract_compressed_file(
    fpath: str,
    extracted_dir_prefix: Optional[str] = None, 
    params: Union[s3.S3Params, fs.LocalFSParams] = None,
    client: Union[s3.Session, None] = None,
    **kwargs
) -> str:
    """Function to extract zip file contents

    Compression Support:
    - .zip
    - .gzip [TODO]
    - .tar.gz [TODO]

    :param fpath: path to compressed file or key to extract
    :param extracted_dir_prefix: directory to prefix the extracted archive, defaults to None
    :param params: either instance of s3.S3Params or fs.LocalFSParams, defaults to None
    :param client: instance of either s3.boto3.Sesssion or None, defaults to None
    :return: path to directory or prefix path where uncompressed file contents exist
    """
    # if no params passed in get default params
    if params is None:  params = get_default_params()
    # create session client if not provided
    if client is None: client = get_session_client(params)

    # handle local fs function
    if isinstance(params, fs.LocalFSParams) and client is None:
        return fs.extract_compressed_file_fs(
            fpath, 
            extracted_dir_prefix=extracted_dir_prefix, 
            local_fs_params=params,
            **kwargs
        )
    
    # handle s3 function
    return s3.extract_compressed_file_s3(
        fpath, 
        extracted_dir_prefix=extracted_dir_prefix,
        s3_params=params,
        s3_client=client,
        **kwargs
    )




from virt_s3.utils import (
    get_custom_logger, 
    ImageFormatType, 
    archive_zip_as_buffer,
    format_bytes
)
from virt_s3.main import (
    get_default_params,
    get_session_client,
    create_bucket,
    get_file_chunked,
    get_file,
    get_image,
    get_files_generator,
    get_files_batch,
    list_dirs,
    get_valid_file_paths,
    file_exists,
    upload_data,
    delete_file,
    delete_files_by_dir,
    extract_compressed_file,
    SessionManager
)
from virt_s3.s3 import S3Params, PredictConnectionStoreParams, TransferConfig
from virt_s3.fs import LocalFSParams
from virt_s3.extras import read_parquet_file_df, write_parquet_file_df

logger = get_custom_logger()


__all__ = [
    'get_custom_logger',
    'PredictConnectionStoreParams',
    'S3Params',
    'SessionManager',
    'TransferConfig',
    'LocalFSParams',
    'ImageFormatType',
    'get_default_params',
    'get_session_client',
    'create_bucket',
    'get_file_chunked',
    'get_file',
    'get_image',
    'get_files_generator',
    'get_files_batch',
    'list_dirs',
    'get_valid_file_paths',
    'file_exists',
    'upload_data',
    'delete_file',
    'delete_files_by_dir',
    'archive_zip_as_buffer',
    'extract_compressed_file',
    'read_parquet_file_df',
    'write_parquet_file_df',
    'format_bytes'
]

__version__ = "0.1.1"


from virt_s3.utils import get_custom_logger, ImageFormatType, archive_zip_as_buffer, KB, MB

from io import BytesIO
from dataclasses import dataclass as pydataclass
from typing import List, Generator, Union, Any, Dict, Optional

import os, shutil, pathlib, zipfile, base64

logger = get_custom_logger()


# Numpy and Pillow
try: 
    from PIL import Image
except ModuleNotFoundError as me:
    class Image:
        class Image: 
            pass
    logger.warning(f"{me} -> `$ poetry install -E image` to use.")

try: 
    import numpy as np
except ModuleNotFoundError as me:
    class np:
        class ndarray:
            pass
    logger.warning(f"{me} -> `$ poetry install -E image` to use.")



@pydataclass
class LocalFSParams:
    """Dataclass for Using Local File System for all S3 Calls
    
    :param use_local_fs: flag to use the local file system or not for S3 calls
    :param root_dir: local file system root dir that represents your 'bucket'
    :param user: your username
    :param bucket_name: emulated S3 'bucket' directory in file system within root dir
    """
    use_local_fs: bool
    root_dir: str
    user: str
    bucket_name: str


class LocalFSSessionManager:
    """Session Context Manager for Local File System

    :param local_fs_params: instance of LocalFSParams, defaults to None
    :param client: existing session client can be passed in, defaults to None
    :param keep_open: boolean flag to keep the session open or not outside of the context manager scope, defaults to False
    """
    def __init__(
        self,
        local_fs_params: LocalFSParams = None,
        client = None,
        keep_open: bool = False,
        **session_client_kwargs
    ) -> None:
        # use default if not provided
        if local_fs_params is None: 
            local_fs_params = get_local_fs_params()
        self.local_fs_params = local_fs_params
        self.client = client
        self.kwargs = session_client_kwargs
        self.keep_open = keep_open
        self.status = None

    def __enter__(self): 
        self.status = "open"
        return self.client

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value or exc_type or traceback:
            logger.error(f"LocalFS Session Error. {exc_type}:{exc_value}\n{traceback}")
        if self.keep_open is False:
            self.status = "closed"


##############################################################
################ Local File System Functions #################
##############################################################


def get_local_fs_params() -> LocalFSParams:
    """Function to get local file system parameters.

    Checks for the following Environment Variables:
    1. LOCAL_FS -> 1 or 0 (indicating whether to use local filesystem or not. 1 is True, 0 is False). Default behavior is False.
    2. LOCAL_FS_ROOT_DIR -> path to local file system root dir
    3. LOCAL_FS_USER -> username (email address) for app you're trying to emulate
    4. S3_DEFAULT_BUCKET -> bucket name proxy for local file system. Defaults to 'default_bucket'.

    :return: instance of LocalFSParams
    """
    logger.debug(f"Getting Local FS Params ...")
    use_local: bool = True
    if os.getenv('LOCAL_FS', "0") == "0": use_local = False
    root_dir: str = os.getenv('LOCAL_FS_ROOT_DIR', os.path.join(os.getcwd(), 'tests/data'))
    user = os.getenv('LOCAL_FS_USER', "default@default.com")
    bucket_name = os.getenv('S3_DEFAULT_BUCKET', 'default-bucket')
    return LocalFSParams(use_local_fs=use_local, root_dir=root_dir, user=user, bucket_name=bucket_name)


def fix_fs_path(
    fpath: str, 
    local_fs_params: LocalFSParams,
) -> str:
    """Function to fix the file or directory path on a local file system
    according to root directory from LocalFSParams. Checks to see if root directory
    is in the path or not and fixes it to be a valid path. Else, it joins the root
    directory with the input path

    :param fpath: input path
    :param local_fs_params: instance of LocalFSParams
    :return: updated file system path
    """
    logger.debug(f"Fixing File Path to Fit Local FS: {fpath}")
    # check if fpath has already been fixed
    if fpath.startswith(os.path.join(local_fs_params.root_dir, local_fs_params.bucket_name)):
        return fpath

    # split by directory
    tmp_path_components = fpath.strip("/").split("/")
    tmp_root_dir = local_fs_params.root_dir.strip("/")

    # find matching path parts - see if root dir in key file path
    # assume root dir in key
    root_dir_in_key, bucket_in_key = True, True
    for component in tmp_root_dir.split("/"):
        if component not in tmp_path_components:
            root_dir_in_key = False
    
    last_root_dir_component = component
    
    # find if bucket already in file path or not
    # assume bucket name in key by default
    if local_fs_params.bucket_name:
        if local_fs_params.bucket_name not in tmp_path_components: bucket_in_key = False
    else: bucket_in_key = False
        
    ## Logic Conditions:
    # 1. root dir in save key & bucket name not in save key
    if root_dir_in_key and not bucket_in_key:
        # new start index = last component of root dir path index + 1
        new_path_start_index = tmp_path_components.index(last_root_dir_component) + 1
        new_path = "/".join(tmp_path_components[new_path_start_index:])
        # if bucket name is in root dir last component
        if local_fs_params.bucket_name in last_root_dir_component:
            return os.path.join(local_fs_params.root_dir, new_path)
        return os.path.join(local_fs_params.root_dir, local_fs_params.bucket_name, new_path)
    # 2. root dir in save key & bucket name in save key
    elif root_dir_in_key and bucket_in_key:
        # get index of last root dir component
        root_dir_end_index = tmp_path_components.index(last_root_dir_component)
        # get index of bucket name
        bucket_index = tmp_path_components.index(local_fs_params.bucket_name)
        # make sure bucket index comes after root dir end
        new_path = "/".join(tmp_path_components[bucket_index+1:])        
        if bucket_index < root_dir_end_index:
            new_path = "/".join(tmp_path_components[root_dir_end_index+1:])
        # if bucket name is in root dir last component
        if local_fs_params.bucket_name in last_root_dir_component:
            return os.path.join(local_fs_params.root_dir, new_path)
        return os.path.join(local_fs_params.root_dir, local_fs_params.bucket_name, new_path)

    # 3. root dir not in save key & bucket name in save key
    elif not root_dir_in_key and bucket_in_key:
        # get index of bucket name
        bucket_index = tmp_path_components.index(local_fs_params.bucket_name)
        new_path = "/".join(tmp_path_components[bucket_index+1:])
        # if bucket name is in root dir last component
        if local_fs_params.bucket_name in last_root_dir_component:
            return os.path.join(local_fs_params.root_dir, new_path)
        return os.path.join(local_fs_params.root_dir, local_fs_params.bucket_name, new_path)

    # 4. root dir not in save key & bucket name not in save key
    # if bucket name is in root dir last component
    if local_fs_params.bucket_name in last_root_dir_component:
        return os.path.join(local_fs_params.root_dir, fpath)
    return os.path.join(local_fs_params.root_dir, local_fs_params.bucket_name, fpath)


def create_fs_bucket(
    bucket_path: str, 
    local_fs_params: LocalFSParams = None,
    **kwargs
) -> bool:
    """Function to create a new 'bucket' directory in local file system

    :param bucket_path: path to new 'bucket' directory
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :return: True if successfully created, else False
    """
    logger.debug(f"Creating Bucket in Local FS: {bucket_path}")
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    bucket_path = fix_fs_path(bucket_path, local_fs_params)

    path = pathlib.Path(bucket_path)
    path.mkdir(exist_ok=True, parents=True)
    if os.path.exists(bucket_path) and os.path.isdir(bucket_path):
        return True
    
    return False
    

def get_fs_file_chunked(
    fpath: str, 
    local_fs_params: LocalFSParams = None,
    chunk_size: int = 1 * MB,
    bytes_io: bool = False,
    **kwargs
) -> Union[BytesIO, bytes, None]: 
    """Function to read a file from local file system in chunking loop

    :param fpath: file path
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :param chunk_size: integer number of bytes per chunk to download, defaults to 1 * MB
    :return: either BytesIO object, bytes object, or None
    """
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    fpath = fix_fs_path(fpath, local_fs_params)

    if not os.path.exists(fpath): return
    # chunking parameters
    
    # resulting buffer to write to
    buffer = BytesIO()

    with open(fpath, 'rb') as f:
        while True:
            # read in chunk bytes
            chunk = f.read(chunk_size)
            # if returned chunk is None, don't process
            if not chunk: break
            # write  chunk to buffer
            buffer.write(chunk)
    
    # seek buffer to start
    buffer.seek(0)
    # return None if empty buffer
    if buffer.getbuffer().nbytes < 1: return None
    
    # bytesio
    if bytes_io: return buffer
    # bytes
    return buffer.getvalue()


def get_fs_file(
    fpath: str, 
    local_fs_params: LocalFSParams = None,
    bytes_io: bool = False,
    **kwargs
) -> Union[BytesIO, bytes, None]:
    """Function to read file from local file system

    :param fpath: path to file
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: instance of bytes or BytesIO object
    """
    logger.debug(f"Getting File from Local FS: {fpath}")
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    fpath = fix_fs_path(fpath, local_fs_params)

    if not os.path.exists(fpath): return
    if bytes_io is False:
        with open(fpath, 'rb') as f:
            data = f.read()
        return data
    
    with open(fpath, 'rb') as f:
        data = f.read()
        data = BytesIO(data)
        data.seek(0)
    
    return data


def get_fs_image(
    fpath: str, 
    img_format: Optional[ImageFormatType] = None,
    local_fs_params: LocalFSParams = None,
    bytes_io: bool = False,
    **kwargs
) -> Union[BytesIO, Image.Image, np.ndarray, str, bytes, None]:
    """Function to get image from local file system

    :param fpath: path to local image file
    :param img_format: oneof ImageFormatType enum, defaults to None
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: an instance of either BytesIO, PIL.Image.Image, np.ndarray, string, or None
    """
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    fpath = fix_fs_path(fpath, local_fs_params)

    ## Format functions
    def _get_base64_html_string(fpath: str) -> str:
        # Get image file
        object_data: BytesIO = get_fs_file(fpath, bytes_io=True)

        # if using base64 encoding
        extension = os.path.splitext(fpath)[1].lower()
        extension_dict = {
            ".jpg": 'data:image/jpeg;base64,',
            ".jpeg": 'data:image/jpeg;base64,',
            ".png": 'data:image/png;base64,',
            ".gif": 'data:image/gif;base64,'
        }
        if extension not in extension_dict: 
            return
        image_string = extension_dict[extension]
        image_string += base64.b64encode(object_data.read()).decode('utf-8')
        return image_string

    def _get_pil_image(fpath: str) -> Image.Image:
        # Get image file
        object_data: BytesIO = get_fs_file(fpath, bytes_io=True)
        with Image.open(object_data) as img:
            img.load()
        return img
    
    def _get_np_array_image(fpath: str) -> np.ndarray:
        # Get image file
        object_data: BytesIO = get_fs_file(fpath, bytes_io=True)
        img = np.array(_get_pil_image(object_data))
        return img

    def _get_bytesio_image(fpath: str) -> BytesIO:
        # Get image file
        object_data: BytesIO = get_fs_file(fpath, bytes_io=True)
        return object_data
    
    # functions dictionary
    format_funcs_dict = {
        ImageFormatType.BASE64_HTML_STRING: _get_base64_html_string,
        ImageFormatType.PIL_IMAGE: _get_pil_image,
        ImageFormatType.NUMPY_ARRAY: _get_np_array_image,
        "bytes_io": _get_bytesio_image
    }

    ## Hierarchy of Logic: `img_format` > `bytes_io`
    # first check for img_format args
    # then check for bytes_io arg
    # if nothing provided return raw output of s3_object.read()
    if img_format in format_funcs_dict: return format_funcs_dict[img_format](fpath)
    if bytes_io: return format_funcs_dict['bytes_io'](fpath)

    return get_fs_file(fpath, bytes_io=False)   


def get_fs_files_generator(
    fpath_list: List[str],
    local_fs_params: LocalFSParams = None, 
    bytes_io: bool = False,
    **kwargs
) -> Generator[Union[BytesIO, bytes, None], Any, None]:
    """Generator function to quickly loop through reading a list of file paths from local file system

    :param fpath_list: list of file paths
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :yield: instance of bytes or BytesIO object
    """
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix paths
    fpath_list = [fix_fs_path(fpath, local_fs_params) for fpath in fpath_list]

    for fpath in fpath_list:
        yield get_fs_file(fpath, bytes_io=bytes_io)

    return


def get_fs_files_batch(
    fpath_list: List[str],
    local_fs_params: LocalFSParams = None,
    max_concurrency: int = 0, 
    bytes_io: bool = False,
    **kwargs
) -> List[Union[BytesIO, bytes, None]]:
    """Function to read a list of file paths from local file system and return a batch of data.

    :param fpath_list: list of file paths
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :param max_concurrency: max no. of concurrent threads to use when downloading multiple files (0 means no concurrency), defaults to 0
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: list of bytes or BytesIO object
    """
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix paths
    fpath_list = [fix_fs_path(fpath, local_fs_params) for fpath in fpath_list]

    # TODO: implement multithreading

    li = []
    for count, fpath in enumerate(fpath_list):
        li.append(get_fs_file(fpath, bytes_io=bytes_io))
        print(f"Completed Downloading: {count+1}/{len(fpath_list)}")

    return li


def list_fs_folders(
    dir_path: str,
    local_fs_params: LocalFSParams = None,
    **kwargs
) -> List[str]:
    """Function to list all 1st level subdirectories in a given directory path

    :param dir_path: input directory path
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :return: list of subdirectory paths
    """
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    dir_path = fix_fs_path(dir_path, local_fs_params)

    ret = []
    for p in os.listdir(dir_path):
        if os.path.isdir(os.path.join(dir_path, p)):
            ret.append(os.path.join(dir_path, p))
    return ret


def get_valid_fs_fpaths(
    dir_path: str,
    ignore: List[str] = [".DS_Store"],
    filter_extensions: List[str] = [],
    local_fs_params: LocalFSParams = None,
    **kwargs
) -> List[str]:
    """Funciton to get list of valid filepaths in file system (recursively) starting from
    and input directory path. This includes nested file paths.

    :param dir_path: input directory path to start from
    :param ignore: list of directories/prefixes and filepaths to ignore, defaults to ['.DS_Store']
    :param filter_extensions: list of file extensions to filter for (empty list means get everything), defaults to []
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :return: list of valid file paths
    """
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    dir_path = fix_fs_path(dir_path, local_fs_params)

    ignore = set(ignore)

    fpaths = []
    for root, dirs, files in os.walk(dir_path, topdown=True): 
        for fname in files:
            if fname.startswith("."): continue
            fpaths.append(os.path.join(root, fname))
    
    # Process keys based on ignore_dirs and filter_extensions
    li = []
    for fpath in fpaths:
        parts = fpath.strip("/").split("/")
        fname = parts[-1]

        # check if ignored
        ignored = False
        for part in parts:
            if part in ignore:
                ignored = True
                break
        
        # check if file extension supported
        # if filter_extensions is empty list -- default to supporting all extensions
        supported_extension = False
        if len(filter_extensions) < 1:
            supported_extension = True
        else:
            for extension in filter_extensions:
                if fname.endswith(extension):
                    supported_extension = True
            
        # only append if not ignored and extension supported
        if not ignored and supported_extension: 
            li.append(fpath)
    
    fpaths, li = li, None
    return fpaths


def fpath_exists_fs(
    fpath: str,
    local_fs_params: LocalFSParams = None,
    **kwargs
) -> bool:
    """Function to check if file path exists

    :param fpath: input file path
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :return: True or False if it exists
    """
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    fpath = fix_fs_path(fpath, local_fs_params)

    return os.path.exists(fpath)


def upload_to_fs(
    data: Union[bytes, memoryview, BytesIO], 
    save_path: str,
    local_fs_params: LocalFSParams = None,
    **kwargs
) -> str:
    """Function to write data to local file system

    :param data: input data to upload
    :param save_path: save path for newly written file
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :return: saved file path
    """
    logger.debug(f"Uploading Data to File: {save_path}")
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    save_path = fix_fs_path(save_path, local_fs_params)

    # if data is BytesIO convert to bytes
    if isinstance(data, BytesIO): data = data.getbuffer()
    if not isinstance(data, (bytes, memoryview)):
        raise Exception(f"UploadToLocalFSError: data is not of instance 'bytes' or 'memoryview' for {save_path}.")
    
    # make directories if not already existing
    path = pathlib.Path(save_path)
    path.parent.mkdir(exist_ok=True, parents=True)
    with open(save_path, 'wb') as f:
        f.write(data)
    
    return save_path


def delete_fs_file(
    fpath: str,
    local_fs_params: LocalFSParams = None,
    **kwargs
) -> bool:
    """Function to delete file from local file system

    :param fpath: input file path to delete
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :return: True if deleted, False if not deleted
    """
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    fpath = fix_fs_path(fpath, local_fs_params)

    if not fpath_exists_fs(fpath): return False
    # delete fpath
    os.remove(fpath)
    # if fpath still exists, it has not been deleted successfully
    if fpath_exists_fs(fpath): return False
    return True


def delete_fs_files_by_dir(
    dir_path: str,
    local_fs_params: LocalFSParams = None,
    **kwargs
) -> List[str]:
    """Function to delete all files and sub-directories in given directory path (including itself)

    :param dir_path: input directory path to delete
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :return: list of deleted filepaths
    """
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    dir_path = fix_fs_path(dir_path, local_fs_params)

    deleted_fpaths: List[str] = get_valid_fs_fpaths(dir_path)
    if len(deleted_fpaths) < 1: return deleted_fpaths
    shutil.rmtree(dir_path)

    for i, fpath in enumerate(deleted_fpaths):
        logger.debug(f"Deleted file: {i+1}/{len(deleted_fpaths)} -- {fpath}")
    return deleted_fpaths


def extract_compressed_file_fs(
    fpath: str, 
    extracted_dir_prefix: Optional[str] = None, 
    local_fs_params: LocalFSParams = None,
    **kwargs
) -> str:
    """Function to extract zip file contents in local file system

    Compression Support: 
    - .zip
    - .gzip [TODO]
    - .tar.gz [TODO]

    :param fpath: path to compressed file to extract
    :param extracted_dir_prefix: directory to prefix the extracted archive, defaults to None
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :return: path to directory where uncompressed file contents exist
    """
    # get default local fs params
    if local_fs_params is None: local_fs_params = get_local_fs_params()
    # fix path
    fpath = fix_fs_path(fpath, local_fs_params)

    def _decompress_zip_and_upload_to_fs(buffer: BytesIO, dir_path: str, zip_fpath: str = None) -> List[str]:
        """Inner function to unzip .zip file in S3 and upload unzipped files
        back to Local FS

        :param buffer: zip file buffer
        :param dir_path: path to parent directory of key .zip file
        :param zip_fpath: path to zip file, defaults to None 
        :return: list of uploaded fpaths
        """
        ret = []
        # new ZipFile instance
        try:
            with zipfile.ZipFile(buffer) as z:
                for i, filename in enumerate(z.namelist()): 
                    # ignore directories and MACOSX specific things
                    if filename.endswith("/") or "MACOSX" in filename: continue
                    logger.debug(f"Extracting File Name: {filename}")
                    # get zipfile info
                    file_info = z.getinfo(filename)
                    # create new save key
                    save_key = os.path.join(dir_path, filename) 
                    # Decompress and copy the files to the 'unzipped' folder 
                    uploaded_key = upload_to_fs(z.open(file_info).read(), save_key)

                    # if nested .zip files, recursively extract
                    if filename.endswith(".zip"):
                        # parse new dir path
                        new_dir_path = uploaded_key.split("/")[:-1]
                        new_dir_path = os.path.join(*new_dir_path)
                        if dir_path.startswith("/") and not new_dir_path.startswith("/"):
                            new_dir_path = f"/{new_dir_path}"
                        # add trailing slash if does not exist
                        if not new_dir_path.endswith("/"):
                            new_dir_path = f"{new_dir_path}/"
                        
                        logger.debug(f"Nested .zip File Found. New Dir Path: {new_dir_path}")
                        buffer = get_fs_file(uploaded_key, bytes_io=True)
                        new_path_li = _decompress_zip_and_upload_to_fs(buffer, new_dir_path, zip_fpath=uploaded_key)
                        ret.extend(new_path_li)
                    else:
                        ret.append(uploaded_key)
        except Exception as e:
            logger.error(f"ERROR! ***DecompressZipFileError*** for file: {zip_fpath}. {e}")
        
        return ret

    # read in compressed file
    buffer = get_fs_file(fpath, bytes_io=True)
    if buffer is None: 
        raise Exception(f"Error Reading File: {fpath} from file system")

    # get last directory path
    path_parts = fpath.split("/")
    dir_path = ""
    fname = path_parts[-1]
    if len(path_parts) > 1: 
        dir_path = os.path.join(*path_parts[:-1])
        if fpath.startswith("/") and not dir_path.startswith("/"):
            dir_path = f"/{dir_path}"
    
    # add extracted dir prefix if passed in
    if extracted_dir_prefix:
        dir_path = os.path.join(dir_path, extracted_dir_prefix)

    # get original file name index in split path
    original_file_index = path_parts.index(fname)
    path_offset = 1
    if extracted_dir_prefix:
        path_offset = 2

    # add trailing slash if does not exist
    if not dir_path.endswith("/"):
        dir_path = f"{dir_path}/"
    
    logger.debug(f"Dirpath: {dir_path}")

    # handle different compressions
    new_keys = []
    if fpath.endswith(".zip"):
        new_keys = _decompress_zip_and_upload_to_fs(buffer, dir_path, zip_fpath=fpath)
    elif fpath.endswith(".gzip"):
        pass
    elif fpath.endswith(".tar.gz"):
        pass

    logger.debug(f"extract_compressed_file_fs() -> New Extracted File Paths: {new_keys[:10]}, ... +{len(new_keys)-10}")

    # get archive dir
    if len(new_keys) < 1:
        logger.warning(f"WARNING: extract_compressed_file_fs() -> resulted in 0 Extracted File Paths!")
        return
    
    # get last directory path
    archive_dirpath = None
    for k in new_keys:
        # split key by path
        new_path_parts = k.split("/")
        # we need split path to have at least the original compressed file level
        if len(new_path_parts) < original_file_index+path_offset: continue
        # format the resulting archive dirpath to return
        archive_dirpath = os.path.join(*new_path_parts[:original_file_index+path_offset])
        if fpath.startswith("/") and not archive_dirpath.startswith("/"):
            archive_dirpath = f"/{archive_dirpath}"
        # make sure archive_dirpath ends with "/"
        if not archive_dirpath.endswith("/"): archive_dirpath =f"{archive_dirpath}/"
        # break if found. We only need one.
        break
    
    # return the archive dir/prefix path
    return archive_dirpath 




from enum import Enum
from typing import Dict
from io import BytesIO

import logging, zipfile

KB = 1024
MB = KB * KB


class ImageFormatType(Enum):
    """Enum class type for Image Format Types

    Enum Options:
    - PIL_IMAGE = "pil_img"
    - BASE64_HTML_STRING = "base64_html_string"
    - NUMPY_ARRAY = "numpy_array"
    """
    PIL_IMAGE = "pil_img"
    BASE64_HTML_STRING = "base64_html_string"
    NUMPY_ARRAY = "numpy_array"


def get_custom_logger(name: str = "VirtS3", log_level: int = logging.DEBUG) -> logging.Logger:
    """Function to return a custom formatted logger object

    :param name: name of logger, defaults to 'VirtS3'
    :param log_level: desired logging level, defaults to logging.DEBUG
    :return: custom formatted Logger streaming to stdout
    """
    logger = logging.getLogger(name)
    logger.propagate = False
    
    if len(logger.handlers) > 0: return logger
    
    # handlers
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(log_level)

    # formatters
    formatter = logging.Formatter(f'[%(name)s][%(filename)s][%(funcName)s][L%(lineno)d][%(asctime)s][%(levelname)s]: %(message)s', "%Y-%m-%d %H:%M:%S")
    stream_handler.setFormatter(formatter)
    logger.setLevel(log_level)
    logger.handlers = [stream_handler]
    return logger


def archive_zip_as_buffer(data_bytes_dict: Dict[str, bytes]) -> BytesIO:
    """Function to create a zip archive from dictionary of expected archive filepaths
    and data bytes
    
    :param data_bytes_dict: { "/path/to/data/file/in/archive/file.extension": data_as_bytes_obj }
    :return: zipped buffer BytesIO object
    """
    zip_buffer = BytesIO()
    with zipfile.ZipFile(zip_buffer, "a", zipfile.ZIP_DEFLATED, False) as z:
        for fpath, data_bytes in data_bytes_dict.items():
            z.writestr(fpath, data_bytes)
    zip_buffer.seek(0)

    return zip_buffer


def format_bytes(num_bytes: int) -> str:
    """Funtion to take as input a number of bytes 
    and return a formatted string for B, KB, MB, GB
    rounded to 2 decimal places

    :param num_bytes: integer number of bytes
    :return: formatted byte string (e.g 5.11 MB )
    """
    # Define the units in increasing size order
    units = ["B", "KB", "MB", "GB", "TB"]
    
    # Start with bytes, and progressively divide by 1024 for each next unit
    unit_index = 0
    while num_bytes >= 1024 and unit_index < len(units) - 1:
        num_bytes /= 1024
        unit_index += 1

    # Format the number to 2 decimal places and return the formatted string
    return f"{num_bytes:.2f} {units[unit_index]}"


from virt_s3.utils import get_custom_logger, ImageFormatType, archive_zip_as_buffer, MB
from concurrent.futures import ThreadPoolExecutor
from io import BytesIO
from dataclasses import dataclass as pydataclass
from typing import List, Generator, Union, Any, Optional, Callable, Dict
from urllib.parse import urlparse
import os, zipfile, base64, time

logger = get_custom_logger()

# Predict backend
try: 
    from predict_backend.store.connection_store import connection_store
except ModuleNotFoundError as me: 
    logger.warning(f"{me} -> Install predict-backend to use.")
    connection_store = None

# Boto3
try:
    from boto3 import Session
    from boto3.s3.transfer import TransferConfig
    from botocore.exceptions import ClientError
except ModuleNotFoundError as me:
    logger.warning(f"{me} -> `$ poetry install -E s3` to use.")
    class Session: 
        pass
    class TransferConfig: 
        def __init__(self, max_concurrency: int = 3):
            pass
    class ClientError(Exception):
        def __init__(self, *args):
            super().__init__(*args)


# Numpy and Pillow
try: 
    from PIL import Image
except ModuleNotFoundError as me:
    class Image:
        class Image: 
            pass
    logger.warning(f"{me} -> `$ poetry install -E image` to use.")

try: 
    import numpy as np
except ModuleNotFoundError as me:
    class np:
        class ndarray:
            pass
    logger.warning(f"{me} -> `$ poetry install -E image` to use.")


@pydataclass
class PredictConnectionStoreParams:
    """Dataclass for Predict Connection Store Parameters

    :param user_id: predict user id
    :param store_name: connection store name
    :param store_owner: connection store owner
    """
    user_id: str
    store_name: str
    store_owner: str

@pydataclass
class S3Params:
    """Dataclass for S3 Boto3 Connection Parameters

    Required:
    - `bucket_name`
    - `endpoint_url`

    Optional:
    - `region_name`

    Authentication Methods (and necessary params):

    1. Explicit Credentials
        - `aws_access_key_id`
        - `aws_secret_access_key`

    2. AWS Profile
        - `profile_name`

    3. Predict Backend Connection Store
        - `connection_store_params`

    :param bucket_name: aws s3 bucket
    :param endpoint_url: aws s3 url
    :param aws_access_key_id: your aws access key id, defaults to None
    :param aws_secret_access_key: your aws secret access key, defaults to None
    :param region_name: your aws region name, defaults to None
    :param profile_name: your aws IAM profile name, defaults to None
    :param connection_store_params: instance of PredictConnectionStoreParams, defaults to None
    """
    # AWS
    bucket_name: str
    endpoint_url: str
    aws_access_key_id: str = None
    aws_secret_access_key: str = None
    region_name: str = None
    profile_name: str = None
    # Predict
    connection_store_params: PredictConnectionStoreParams = None


class S3SessionManager:
    """Session Context Manager for S3

    :param s3_params: instance of S3Params, defaults to None
    :param s3_client: boto3 s3 client object, defaults to None
    :param keep_open: boolean flag to keep the session open or not outside of the context manager scope, defaults to False
    """
    def __init__(
        self,
        s3_params: S3Params = None,
        s3_client: Session = None,
        keep_open: bool = False,
        **session_client_kwargs
    ) -> None:

        # use default if not provided
        if s3_params is None and s3_client is None: 
            s3_params = get_s3_default_params()
        self.s3_params = s3_params
        self.s3_client = s3_client
        self.kwargs = session_client_kwargs
        self.keep_open = keep_open
        self.status = None

    def __enter__(self): 
        # create session client if not provided
        if self.s3_client is None: 
            self.s3_client = get_s3_session_client(self.s3_params, **self.kwargs)
        self.status = "open"
        return self.s3_client

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value or exc_type or traceback:
            logger.error(f"S3 Session Error. {exc_type}:{exc_value}\n{traceback}")

        if self.s3_client and self.keep_open is False:
            self.s3_client.close()
            self.status = "closed"        


# Valid S3 Session and Client Args
VALID_S3_SESSION_ARGS = {
    'aws_access_key_id',
    'aws_secret_access_key',
    'aws_session_token',
    'region_name',
    'botocore_session',
    'profile_name'
}
VALID_S3_CLIENT_ARGS = {
    'region_name',
    'api_version',
    'use_ssl',
    'verify',
    'endpoint_url',
    'aws_access_key_id',
    'aws_secret_access_key',
    'aws_session_token',
    'config'
}


## Manage in-memory record of which function is fastest for upload and download over time
# calculating megabytes per second
# Download: get_s3_file_simple() or get_s3_file_transfer()
# Upload: upload_to_s3_simple() or upload_to_s3_transfer()

MAX_UPLOAD_DOWNLOAD_COUNT = 10

@pydataclass
class PerformanceMetrics:
    """Dataclass for different performance metrics

    :param mbps: megabytes per second, defaults to 0.0
    """
    mbps: float = 0.0

class PerformanceRecord:
    """Encpasulates performance for a given performance monitoring type

    :param _type: 'upload' or 'download'
    :param metrics: dictionary of function name str -> list of PerformanceMetrics, defaults to { 'simple': [], 'transfer': [] }
    :param count: number of times this performance record has been updated, defaults to 0
    :param chosen_func: the specific function key name that was chosen to be run, defaults to None
    """
    def __init__(
        self,
        _type: str,
        metrics: Dict[str, List[PerformanceMetrics]] = {
            'simple': [],
            'transfer': []
        },
        count: int = 0,
        chosen_func: str = None
    ) -> None:
        self._type = _type
        self.metrics = metrics
        self.count = count
        self.chosen_func = chosen_func
   

class UploadDownloadPerformanceManager:
    """Class used to record and determine which different upload or download
    function to run based on a record of performance metrics over the course of
    the program run lifetime
    """
    def __init__(self) -> None:
        self.records_dict = {
            'upload': PerformanceRecord('upload'),
            'download': PerformanceRecord('download')
        }

    def reset(self, _type: str):
        """Method to reset the performance records dictionary

        :param _type: 'upload' or 'download'
        """
        self.records_dict[_type].metrics['simple'] = []
        self.records_dict[_type].metrics['transfer'] = []
        self.records_dict[_type].count = 0

    def average_metrics(self, metrics_li: List[PerformanceMetrics]) -> PerformanceMetrics:
        """Method to get average of list of metrics

        :param metrics_li: list of PerformanceMetrics
        :return: single instance of PerformanceMetrics aggregated as simple average
        """
        n, total = len(metrics_li), 0.0
        if n < 1: return PerformanceMetrics(mbps=0.0)
        for metric in metrics_li: total += metric.mbps
        avg = total / n
        return PerformanceMetrics(mbps=round(avg, 2))

    def get_metrics(self, _type: str, func_str: str) -> List[PerformanceMetrics]:
        """Method to get the specific function PerformanceMetrics

        :param _type: 'upload' or 'download'
        :param func_str: name of function key that you want to return list of PerformanceMetrics for
        :return: list of PerformanceMetrics
        """
        return self.records_dict[_type].metrics[func_str]
        
    def choose_function(self, _type: str) -> str:
        """Method to choose the right function to run based on historical throughput MBPS

        :param _type: 'upload' or 'download'
        :return: chosen function string name
        """
        # determine which function to use for download
        perf_record = self.records_dict[_type]
        simple_mbps = self.average_metrics(perf_record.metrics['simple']).mbps
        transfer_mbps = self.average_metrics(perf_record.metrics['transfer']).mbps
        
        # default to 'transfer' option
        chosen_func = "transfer"
        # both no history, use transfer
        if simple_mbps == 0 and transfer_mbps == 0: chosen_func = "transfer"
        # simple has no history, but transfer has history
        elif simple_mbps == 0 and transfer_mbps > 0: chosen_func = "simple"
        # simple has history, but transfer has no history
        elif simple_mbps > 0 and transfer_mbps == 0: chosen_func = "transfer"
        # both have history -- calculate megabytes per second and compare
        elif simple_mbps > 0 and transfer_mbps > 0:
            chosen_func = "simple"
            # compare head to head
            if transfer_mbps > simple_mbps:
                chosen_func = "transfer"
            # Every N calls, reset things
            if perf_record.count >= MAX_UPLOAD_DOWNLOAD_COUNT: 
                logger.debug(f"Shaking Things up! Resetting S3 Download Performance Dictionary ...")
                self.reset(_type)

        # set chosen func
        self.records_dict[_type].chosen_func = chosen_func
        return chosen_func
    
    def update_performance_record(
        self, 
        _type: str, 
        bytes_transferred: int, 
        transfer_time: float
    ) -> None:
        """Method used to update the performance record based
        with throughput metrics

        :param _type: 'upload' or 'download'
        :param bytes_transferred: number of bytes transferred
        :param transfer_time: total seconds the transfer took
        """
        mbps = round((bytes_transferred / MB) / transfer_time, 2)
        chosen_func = self.records_dict[_type].chosen_func
        self.records_dict[_type].metrics[chosen_func].append(PerformanceMetrics(mbps=mbps))
        self.records_dict[_type].count += 1


performance_manager = UploadDownloadPerformanceManager()

###################################################
################ AWS S3 Functions #################
###################################################


def get_s3_default_params() -> S3Params:
    """Function to get kwargs necessary for S3 connections

    :return: instance of S3Params
    """
    logger.debug(f"Getting S3 Params ...")
    # get necessary environment variables
    s3_url = os.environ.get("S3_URL")
    s3_default_bucket = os.environ.get("S3_DEFAULT_BUCKET", "default-bucket")
    aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY")
    aws_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID")
    aws_region = os.environ.get("AWS_REGION")

    kwargs = {
        'bucket_name': s3_default_bucket,
        'endpoint_url': s3_url,
        'region_name': aws_region
    }
    use_access_key = aws_access_key_id is not None and aws_secret_access_key is not None
    if use_access_key:
        access_kwargs = {
            'aws_access_key_id': aws_access_key_id,
            'aws_secret_access_key': aws_secret_access_key
        }
        kwargs = dict(**access_kwargs, **kwargs)
    
    return S3Params(**kwargs)


def get_s3_session_client(s3_params: S3Params, **kwargs) -> Session:
    """Function to get an S3 Boto3 Session client

    Authentication priority in case multiple options are provided:
    1. explicit supplied credentials (aws_access_key_id, aws_secret_access_key)
    2. aws profile
    3. connection store

    :param s3_params: instance of S3Params outlining s3 connectin parameters
    :return: instance of boto3.Session.client that can be used to make requests to S3
    """
    # Format S3 Session Client Args
    s3_args = kwargs

    if s3_params.aws_access_key_id:
        s3_args = {
            "aws_access_key_id": s3_params.aws_access_key_id,
            "aws_secret_access_key": s3_params.aws_secret_access_key,
            **kwargs
        }

    elif s3_params.connection_store_params and connection_store:
        s3_args = {
            **connection_store.get_connection(
                name=s3_params.connection_store_params.store_name,
                connection_owner=s3_params.connection_store_params.store_owner,
                user_id=s3_params.connection_store_params.user_id,
            ),
            **kwargs
        }
    elif s3_params.profile_name:
        s3_args = {
            "profile_name": s3_params.profile_name,
            **kwargs
        }

    # Add S3 URL to all configurations
    if s3_params.endpoint_url:
        s3_args['endpoint_url'] = s3_params.endpoint_url

    session_args = {k: v for k, v in s3_args.items() if k in VALID_S3_SESSION_ARGS}
    client_args = {k: v for k, v in s3_args.items() if k in VALID_S3_CLIENT_ARGS}

    # Get session client
    # config = Config(read_timeout=300, connect_timeout=300, retries={"max_attempts": 3})
    return Session(**session_args).client('s3', **client_args)


def create_s3_bucket(
    bucket_name: str, 
    s3_client = None,
    s3_params: S3Params = None,
    **kwargs
) -> bool:
    """Function to create an S3 Bucket

    :param bucket_name: name of bucket to create
    :param s3_client: boto3 s3 client object, defaults to None
    :param s3_params: instance of S3Params, defaults to None
    :param local_fs_params: instance of LocalFSParams, defaults to None
    :return: True if successfully created, else False
    """
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)
    
    logger.info(f"Creating s3 bucket: {bucket_name} ...")

    # S3 kwargs (defaults are empty, but can be passed in)
    create_args = { "Bucket": bucket_name }
    bucket_encryption_args = {}
    block_bucket_public_access_args = {}
    bucket_policy_args = {}

    # defaults vs passed in
    create_args = kwargs.get('create_args', create_args)
    bucket_encryption_args = kwargs.get('bucket_encryption_args', bucket_encryption_args)
    block_bucket_public_access_args = kwargs.get('block_bucket_public_access_args', block_bucket_public_access_args)
    bucket_policy_args = kwargs.get('bucket_policy_args', bucket_policy_args)
    
    if s3_params.region_name and s3_params.region_name != 'us-east-1':
        create_args['CreateBucketConfiguration'] = {'LocationConstraint': s3_params.region_name}
    try:
        response = s3_client.create_bucket(**create_args)
        logger.info(f'response: {response}')
        logger.info(f'Bucket: {bucket_name} created in {s3_params.region_name}')
    except (s3_client.exceptions.BucketAlreadyExists, s3_client.exceptions.BucketAlreadyOwnedByYou) as e:
        logger.info(f'bucket: {bucket_name} already exists: {e}')
        return True
    except Exception as e:
        logger.error(f'Unhandled Exception in S3Handler.create_bucket: {e}')
        return False
    try:
        if len(bucket_encryption_args) > 0:
            s3_client.put_bucket_encryption(**bucket_encryption_args)
            logger.info(f'Bucket encryption enabled')
        if len(block_bucket_public_access_args) > 0:
            s3_client.put_public_access_block(**block_bucket_public_access_args)
            logger.info(f'Bucket public access blocked')
        if len(bucket_policy_args) > 0:
            s3_client.put_bucket_policy(**bucket_policy_args)
            logger.info(f'Bucket access policy set')
    except Exception as e:
        logger.error(f'Unhandled Exception in S3Handler setting policies: {e}')
    return True


def get_s3_file_chunked(
    s3_fpath: str, 
    s3_client = None,
    s3_params: S3Params = None,
    chunk_size: int = 1 * MB,
    progress_callback: Callable[[int, int], any] = None,
    bytes_io: bool = False,
    **kwargs
) -> Union[BytesIO, bytes, None]: 
    """Function to read an s3 file using a chunking loop. This can be useful
    when trying to get very large files (relies on boto3.s3.get_object())

    :param s3_fpath: s3 key
    :param s3_client: boto3 s3 client object, defaults to None
    :param s3_params: instance of S3Params, defaults to None
    :param chunk_size: integer number of bytes per chunk to download, defaults to 1 * MB
    :param progress_callback: user-defined custom progress callback that is called throughout the download process that takes as input \
        (current number of bytes downloaded so far, total number of bytes to download)
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: either BytesIO object, bytes object, or None
    """
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    # get size of object to read in
    object_size = s3_client.get_object_attributes(
        Bucket=s3_params.bucket_name, 
        Key=s3_fpath, 
        ObjectAttributes=["ObjectSize"]
    ).get("ObjectSize")

    # chunking parameters
    # 100 MB
    chunk_start = 0
    chunk_end = chunk_start + chunk_size - 1

    # handle progress callback
    curr_transferred = 0
    def _custom_progress_callback(chunk_length: int):
        nonlocal curr_transferred
        nonlocal object_size
        curr_transferred += chunk_length
        # call passed in progress callback function
        progress_callback(curr_transferred, object_size)

    # resulting buffer to write to
    buffer = BytesIO()

    while chunk_start <= object_size:
        body = s3_client.get_object(
            Bucket=s3_params.bucket_name, 
            Key=s3_fpath, 
            Range=f"bytes={chunk_start}-{chunk_end}"
        ).get("Body")

        # if returned body is None, don't process
        if not body: continue

        # read in chunk bytes
        chunk = body.read()
        
        # write  chunk to buffer
        buffer.write(chunk)

        # progress callback
        if progress_callback: _custom_progress_callback(chunk_end-chunk_start)
        
        # move iteration forward
        chunk_start += chunk_size
        chunk_end += chunk_size
    
    # seek buffer to start
    buffer.seek(0)

    # return None if empty buffer
    if buffer.getbuffer().nbytes < 1: 
        return None
    
    # bytesio
    if bytes_io: return buffer
    # bytes
    return buffer.getvalue()


def get_s3_file_simple(
    s3_fpath: str, 
    s3_client = None,
    s3_params: S3Params = None,
    bytes_io: bool = False,
    **kwargs
) -> Union[BytesIO, bytes, None]:
    """Function to downlad a given s3 file (relies on boto3.s3.get_object())

    :param s3_fpath: key to s3 file to download
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: either bytes object, BytesIO object, or None
    """
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    try:
        resp = s3_client.get_object(Bucket=s3_params.bucket_name, Key=s3_fpath)
    except ClientError as ce:
        logger.debug(f"Failed to get S3 object: {s3_fpath}")
        return None
    
    # Only continue processing if HTTP 200 Status Code in Repsonse
    if resp.get('ResponseMetadata', {}).get('HTTPStatusCode', 400) != 200: return None
    
    obj = resp['Body']
    
    # Try reading data all at once
    object_data = obj.read()
    # bytes
    if not bytes_io: return object_data
    # BytesIO
    object_data = BytesIO(object_data)
    object_data.seek(0)
    return object_data


def get_s3_file_transfer(
    s3_fpath: str, 
    s3_client = None,
    s3_params: S3Params = None,
    progress_callback: Callable[[int, int], any] = None,
    transfer_config: TransferConfig = TransferConfig(max_concurrency=3),
    bytes_io: bool = False,
    **kwargs
) -> Union[BytesIO, bytes, None]:
    """Function to downlad a given s3 file (relies on boto3.s3.download_fileobj())

    :param s3_fpath: key to s3 file to download
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :param progress_callback: user-defined custom progress callback that is called throughout the download process that takes as input \
        (current number of bytes downloaded so far, total number of bytes to download)
    :param transfer_config: instance of \
        [boto3.s3.transfer.TransferConfig](https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.TransferConfig),\
        defaults to boto3.s3.transfer.TransferConfig(max_concurrency=3)
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: either bytes object, BytesIO object, or None
    """
    # handle progress callback
    callback = None
    if progress_callback:        
        try:
            # get size of object to read in
            object_size = s3_client.get_object_attributes(
                Bucket=s3_params.bucket_name, 
                Key=s3_fpath, 
                ObjectAttributes=["ObjectSize"]
            ).get("ObjectSize")
        except Exception as e:
            logger.debug(f"Encountered Error Trying to Get Object Total Size for {s3_fpath}. {e}")
            object_size = None

        # tracking download progress
        curr_transferred = 0
        def _custom_progress_callback(chunk_size: int):
            nonlocal curr_transferred
            nonlocal object_size
            curr_transferred += chunk_size
            # call passed in progress callback function
            progress_callback(curr_transferred, object_size)

        callback = _custom_progress_callback

    # default return data to None
    buffer = BytesIO()
    s3_client.download_fileobj(
        s3_params.bucket_name, 
        s3_fpath, 
        buffer, 
        Callback=callback,
        Config=transfer_config
    )
    buffer.seek(0)
    # BytesIO
    if bytes_io: data, buffer = buffer, None
    # bytes
    else: data, buffer = buffer.getvalue(), None

    return data


def get_s3_file( 
    s3_fpath: str, 
    s3_client = None,
    s3_params: S3Params = None,
    progress_callback: Callable[[int, int], any] = None,
    transfer_config: TransferConfig = TransferConfig(max_concurrency=3),
    bytes_io: bool = False,
    **kwargs
) -> Union[BytesIO, bytes, None]:
    """Function to downlad a given s3 file using either: `get_s3_file_simple()` or `get_s3_file_transfer()`.
    Falls back to `get_s3_file_chunked()` if either of the above functions above fail.

    :param s3_fpath: key to s3 file to download
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :param progress_callback: user-defined custom progress callback that is called throughout the download process that takes as input \
        (current number of bytes downloaded so far, total number of bytes to download)
    :param transfer_config: instance of \
        [boto3.s3.transfer.TransferConfig](https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.TransferConfig),\
        defaults to boto3.s3.transfer.TransferConfig(max_concurrency=3)
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: either bytes object, BytesIO object, or None
    """
    global performance_manager

    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    # default return data to None
    data = None
    try:
        # start download
        start = time.time()
        perf_type = "download"

        # choose the right function to run
        chosen_func = performance_manager.choose_function(perf_type)
        simple_mbps = performance_manager.average_metrics(performance_manager.get_metrics(perf_type, 'simple')).mbps
        transfer_mbps = performance_manager.average_metrics(performance_manager.get_metrics(perf_type, 'transfer')).mbps
        # call appropriate function
        if chosen_func == "simple":
            logger.debug(f"Simple={simple_mbps}mbps vs Transfer={transfer_mbps}mbps. Calling `get_s3_file_simple()`...")
            data = get_s3_file_simple(s3_fpath, s3_params=s3_params, s3_client=s3_client, bytes_io=bytes_io, **kwargs)
        elif chosen_func == "transfer":
            logger.debug(f"Simple={simple_mbps}mbps vs Transfer={transfer_mbps}mbps. Calling `get_s3_file_transfer()`...")
            data = get_s3_file_transfer(s3_fpath, s3_params=s3_params, s3_client=s3_client, 
                progress_callback=progress_callback, transfer_config=transfer_config, bytes_io=bytes_io, **kwargs)
        
        # update performance dict
        total_seconds = time.time() - start
        n_bytes = 0.0
        if isinstance(data, BytesIO): n_bytes = data.getbuffer().nbytes
        elif isinstance(data, (bytes, memoryview)): n_bytes = len(data)
        performance_manager.update_performance_record(perf_type, n_bytes, total_seconds)

    except Exception as e:
        logger.debug(f"Caught Exception: {e}")
        logger.debug(f"Failed to read S3 object: {s3_fpath}. Falling back to reading in chunks ...")
        logger.debug(f"Attempting to read S3 File in Chunks from {s3_fpath}")
        data = get_s3_file_chunked(
            s3_fpath, 
            s3_params=s3_params,
            progress_callback=progress_callback,
            bytes_io=bytes_io
            **kwargs
        )
    # logger.debug(f"Download Time for {s3_fpath}: {round(total_seconds, 3)}")
    return data
    

def get_s3_image(
    s3_fpath: str, 
    s3_client = None,
    s3_params: Optional[S3Params] = None,
    img_format: Optional[ImageFormatType] = None,
    bytes_io: bool = False,
    **kwargs 
) -> Union[BytesIO, Image.Image, np.ndarray, str, bytes, None]:
    """Function to retrieve an image from S3. 
    Calls `get_s3_file()` so you can also pass in additional kwargs through to that function.

    Formatting Hierarchy Logic: `img_format` > `bytes_io`
    1. Check for img_format args
    2. Then check for bytes_io arg
    3. If neither provided return raw output of s3_object.read()

    :param s3_fpath: key to s3 file to download
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :param img_format: one of ImageFormatType enum, defaults to None
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: an instance of either BytesIO, PIL.Image.Image, np.ndarray, string, bytes, or None
    """
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    ## Format functions
    def _get_base64_html_string(s3_fpath: str) -> str:
        # Get image file
        object_data: BytesIO = get_s3_file(
            s3_fpath, 
            s3_client=s3_client, 
            s3_params=s3_params,
            bytes_io=True,
            **kwargs
        )
        # if using base64 encoding
        extension = os.path.splitext(s3_fpath)[1].lower()
        extension_dict = {
            ".jpg": 'data:image/jpeg;base64,',
            ".jpeg": 'data:image/jpeg;base64,',
            ".png": 'data:image/png;base64,',
            ".gif": 'data:image/gif;base64,'
        }
        if extension not in extension_dict: 
            return
        image_string = extension_dict[extension]
        image_string += base64.b64encode(object_data.read()).decode('utf-8')
        return image_string

    def _get_pil_image(s3_fpath: str) -> Image.Image:
        # Get image file
        object_data: BytesIO = get_s3_file(
            s3_fpath, 
            s3_client=s3_client, 
            s3_params=s3_params,
            bytes_io=True,
            **kwargs
        )
        with Image.open(object_data) as img:
            img.load()
        return img
    
    def _get_np_array_image(s3_fpath: str) -> np.ndarray:
        # Get image file
        object_data: BytesIO = get_s3_file(
            s3_fpath, 
            s3_client=s3_client, 
            s3_params=s3_params,
            bytes_io=True,
            **kwargs
        )
        img = np.array(_get_pil_image(object_data))
        return img

    def _get_bytesio_image(s3_fpath: str) -> BytesIO:
        # Get image file
        object_data: BytesIO = get_s3_file(
            s3_fpath, 
            s3_client=s3_client, 
            s3_params=s3_params,
            bytes_io=True,
            **kwargs
        )
        return object_data
    
    # functions dictionary
    format_funcs_dict = {
        ImageFormatType.BASE64_HTML_STRING: _get_base64_html_string,
        ImageFormatType.PIL_IMAGE: _get_pil_image,
        ImageFormatType.NUMPY_ARRAY: _get_np_array_image,
        "bytes_io": _get_bytesio_image
    }

    ## Hierarchy of Logic: `img_format` > `bytes_io`
    # first check for img_format args
    # then check for bytes_io arg
    # if nothing provided return raw output of s3_object.read()
    if img_format in format_funcs_dict: return format_funcs_dict[img_format](s3_fpath)
    if bytes_io: return format_funcs_dict['bytes_io'](s3_fpath)

    return get_s3_file(s3_fpath, bytes_io=False, **kwargs)


def get_s3_files_generator(
    key_list: List[str], 
    s3_client = None,
    s3_params: S3Params = None,
    bytes_io: bool = False,
    **kwargs
) -> Generator[Union[BytesIO, bytes, None], Any, None]:
    """Generator function to quickly loop through reading a list of s3 file keys.
    Calls `get_s3_file()` so you can also pass in additional kwargs through to that function.

    :param key_list: list of s3 keys
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: None
    :yield: either BytesIO object, bytes object, or None
    """
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    for key in key_list: 
        yield get_s3_file(key, s3_client=s3_client, s3_params=s3_params, bytes_io=bytes_io, **kwargs)
    return


def get_s3_files_batch(
    key_list: List[str],
    s3_client = None,
    s3_params: S3Params = None, 
    max_concurrency: int = 0, 
    bytes_io: bool = False,
    **kwargs
) -> list:
    """Function to read a list of s3 file keys and return a batch data. Can use multithreading as well.
    Calls `get_s3_file()` when `max_concurrency=0` so you can also pass in additional kwargs through to that function.

    :param key_list:  list of s3 keys
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :param max_concurrency: max no. of concurrent threads to use when downloading multiple files (0 means no concurrency), defaults to 0
    :param bytes_io: flag to convert pulled object as BytesIO instance, defaults to False
    :return: either pulled S3 object or BytesIO object
    """
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    global count
    count = 0

    def _download_file(s3_client, s3_params: S3Params, s3_fpath: str):
        global count
        obj = s3_client.get_object(Bucket=s3_params.bucket_name, Key=s3_fpath)
        obj = obj['Body'].read()
        if bytes_io: 
            obj = BytesIO(obj)
            obj.seek(0)
        logger.debug(f"Completed Downloading from S3 (in batch): {count+1}/{len(key_list)}")
        count +=1
        return obj
    
    # Iterative batch mode
    if not max_concurrency or max_concurrency < 1:
        objs = []
        for k in key_list: 
            objs.append(
                get_s3_file(k, s3_params=s3_params, s3_client=s3_client, bytes_io=bytes_io, **kwargs)
            )
        return objs
    
    # Multithreading batch mode
    args = [(a,) for a in key_list]
    objs = []
    with ThreadPoolExecutor(max_workers=max_concurrency) as executor:
        for obj in executor.map(lambda p: _download_file(s3_client, s3_params, *p), args):
            objs.append(obj)
    
    return objs 


def list_s3_folders(
    prefix: str,
    s3_client = None,
    s3_params: S3Params = None,
    delimiter: str ='/',
    **kwargs
) -> List[str]:
    """Function to get list of valid folders within S3 Bucket Prefix
    
    :param prefix: S3 Bucket Path Prefix
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :param delimiter: delimiter used to filter objects in S3 (e.g. only subfolders - delimiter="/"), defaults to '/'
    :return: List of S3 Object Keys
    """
    # Adding trailing slash if looking for "directory"
    if prefix.strip() != "" and not prefix.endswith("/"):
        old_prefix = prefix
        prefix = f"{prefix}/"
        logger.debug(f"Old Prefix {old_prefix} -> New Prefix: {prefix}")

    folders = []
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    paginator = s3_client.get_paginator('list_objects_v2')
    page_iterator = paginator.paginate(
        Bucket=s3_params.bucket_name,
        Prefix=prefix,
        Delimiter=delimiter, 
        PaginationConfig={"PageSize": 100}
    )
    for page in page_iterator:
        if "CommonPrefixes" not in page: continue
        prefixes = [pref_dict['Prefix'] for pref_dict in page['CommonPrefixes']]
        folders.extend(prefixes)
    return folders


def get_valid_s3_keys(
    prefix: str,
    ignore: List[str] = [".DS_Store"],
    filter_extensions: List[str] = [],
    s3_client = None,
    s3_params: S3Params = None,
    **kwargs
) -> List[str]:
    """Function to get list of valid keys within S3 Bucket Path
    
    :param prefix: S3 Bucket Path Prefix
    :param ignore: list of directories/prefixes and filepaths to ignore, defaults to ['.DS_Store']
    :param filter_extensions: list of file extensions to filter for (empty list means get everything), defaults to []
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :return: List of S3 Object Keys
    """
    # Adding trailing slash if looking for "directory"
    if prefix.strip() != "" and not prefix.endswith("/"):
        old_prefix = prefix
        prefix = f"{prefix}/"
        logger.debug(f"Old Prefix {old_prefix} -> New Prefix: {prefix}")
    
    ignore = set(ignore)
    
    valid_keys = []
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    bucket_kwargs = {
        "Bucket": s3_params.bucket_name, 
        "Prefix": prefix,
    }
    while True:
        resp = s3_client.list_objects_v2(**bucket_kwargs)
        for obj in resp["Contents"]: valid_keys.append(obj["Key"])
        if "NextContinuationToken" not in resp: break
        bucket_kwargs['ContinuationToken'] = resp["NextContinuationToken"]
    
    # Process keys based on ignore_dirs and filter_extensions
    li = []
    for fpath in valid_keys:
        parts = fpath.strip("/").split("/")
        fname = parts[-1]

        # check if ignored
        ignored = False
        for part in parts:
            if part in ignore:
                ignored = True
                break
        
        # check if file extension supported
        # if filter_extensions is empty list -- default to supporting all extensions
        supported_extension = False
        if len(filter_extensions) < 1:
            supported_extension = True
        else:
            for extension in filter_extensions:
                if fname.endswith(extension):
                    supported_extension = True
            
        # only append if not ignored and extension supported
        if not ignored and supported_extension: 
            li.append(fpath)
    
    fpaths, li = li, None
    return fpaths


def key_exists_s3(
    key: str, 
    s3_client = None, 
    s3_params: S3Params = None,
    **kwargs
) -> bool:
    """Function to see if key exists in S3

    :param key: key fpath string
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :return: True if key exists, False otherwise
    """
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    try:
        s3_client.head_object(Bucket=s3_params.bucket_name, Key=key)
        return True
    except ClientError as ce:
        if ce.response["Error"]["Code"] == "404":
            return False
        raise


def upload_to_s3_simple(
    data: Union[bytes, memoryview, BytesIO], 
    save_key: str, 
    s3_client = None, 
    s3_params: S3Params = None,
    **kwargs
) -> str:
    """Function to upload data object to s3 using `boto3.s3.put_object()`

    :param data: data to save as bytes, memoryview, or BytesIO
    :param save_key: s3 key to save the data to
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :raises Exception: General exception if not HTTP Status Code 200
    :return: S3 key of uploaded object if successful
    """
    # if data is BytesIO convert to bytes
    if isinstance(data, BytesIO): data = data.getbuffer()
    if not isinstance(data, (bytes, memoryview)):
        raise Exception(f"UploadToS3Error: data is not of instance 'bytes' or 'memoryview' for {save_key}.")
    
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)
    
    # upload
    response = s3_client.put_object(
        Body=data,
        Bucket=s3_params.bucket_name,
        Key=save_key
    )
    if response.get('ResponseMetadata', {}).get('HTTPStatusCode', 400) == 200:
        return save_key

    raise Exception(f'Error writing object to S3')


def upload_to_s3_transfer(
    data: Union[bytes, memoryview, BytesIO], 
    save_key: str, 
    s3_client = None, 
    s3_params: S3Params = None,
    progress_callback: Callable[[int, int], any] = None,
    transfer_config: TransferConfig = TransferConfig(max_concurrency=3),
    **kwargs
) -> str:
    """Function to upload data object to s3 using `boto3.s3.upload_fileobj()`

    :param data: data to save as bytes, memoryview, or BytesIO
    :param save_key: s3 key to save the data to
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :param progress_callback: user-defined custom progress callback that is called throughout the download process that takes as input \
        (current number of bytes uploaded so far, total number of bytes to download)
    :param transfer_config: instance of \
        [boto3.s3.transfer.TransferConfig](https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.TransferConfig),\
        defaults to boto3.s3.transfer.TransferConfig(max_concurrency=3)
    :return: S3 key of uploaded object if successful
    """
    global upload_download_performance_dict

    # if data is BytesIO convert to bytes
    if isinstance(data, BytesIO): data = data.getbuffer()
    if not isinstance(data, (bytes, memoryview)):
        raise Exception(f"UploadToS3Error: data is not of instance 'bytes' or 'memoryview' for {save_key}.")
    
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    # handle progress callback
    callback = None
    if progress_callback:        
        # get size of object to upload
        object_size = len(data)
        # tracking download progress
        curr_transferred = 0
        def _custom_progress_callback(chunk_size: int):
            nonlocal curr_transferred
            nonlocal object_size
            curr_transferred += chunk_size
            # call passed in progress callback function
            progress_callback(curr_transferred, object_size)

        callback = _custom_progress_callback
    
    # call s3 upload
    s3_client.upload_fileobj(
        BytesIO(data),
        s3_params.bucket_name,
        save_key,
        Callback=callback,
        Config=transfer_config
    )

    return save_key


def upload_to_s3(
    data: Union[bytes, memoryview, BytesIO], 
    save_key: str, 
    s3_client = None, 
    s3_params: S3Params = None,
    progress_callback: Callable[[int, int], any] = None,
    transfer_config: TransferConfig = TransferConfig(max_concurrency=3),
    **kwargs
) -> str:
    """Function to upload data object to s3 using either: `upload_to_s3_simple()` or `upload_to_s3_transfer()`.

    :param data: data to save as bytes, memoryview, or BytesIO
    :param save_key: s3 key to save the data to
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :param progress_callback: user-defined custom progress callback that is called throughout the download process that takes as input \
        (current number of bytes uploaded so far, total number of bytes to download)
    :param transfer_config: instance of \
        [boto3.s3.transfer.TransferConfig](https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.TransferConfig),\
        defaults to boto3.s3.transfer.TransferConfig(max_concurrency=3)
    :raises Exception: General exception if not HTTP Status Code 200 or not of right type
    :return: S3 key of uploaded object if successful
    """
    global performance_manager

    # if data is BytesIO convert to bytes
    if isinstance(data, BytesIO): data = data.getbuffer()
    if not isinstance(data, (bytes, memoryview)):
        raise Exception(f"UploadToS3Error: data is not of instance 'bytes' or 'memoryview' for {save_key}.")
    
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)
    
    # default return
    ret = None
    try:
        # start upload
        start = time.time()
        perf_type = "upload"

        # choose the right function to run
        chosen_func = performance_manager.choose_function(perf_type)
        simple_mbps = performance_manager.average_metrics(performance_manager.get_metrics(perf_type, 'simple')).mbps
        transfer_mbps = performance_manager.average_metrics(performance_manager.get_metrics(perf_type, 'transfer')).mbps
        # call appropriate function
        if chosen_func == "simple":
            logger.debug(f"Simple={simple_mbps}mbps vs Transfer={transfer_mbps}mbps. Calling `upload_to_s3_simple()`...")
            ret = upload_to_s3_simple(data, save_key, s3_params=s3_params, s3_client=s3_client, **kwargs)
        elif chosen_func == "transfer":
            logger.debug(f"Simple={simple_mbps}mbps vs Transfer={transfer_mbps}mbps. Calling `upload_to_s3_transfer()`...")
            ret = upload_to_s3_transfer(data, save_key, s3_params=s3_params, s3_client=s3_client, 
                progress_callback=progress_callback, transfer_config=transfer_config, **kwargs)
        
        # update performance dict
        total_seconds = time.time() - start
        n_bytes = 0.0
        if isinstance(data, BytesIO): n_bytes = data.getbuffer().nbytes
        elif isinstance(data, (bytes, memoryview)): n_bytes = len(data)
        performance_manager.update_performance_record(perf_type, n_bytes, total_seconds)

    except Exception as e:
        logger.debug(f"Encountered Error Trying to Upload Object: {save_key}. {e}")

    return ret
    

def delete_s3_object(
    key: str, 
    s3_client = None, 
    s3_params: S3Params = None,
    **kwargs
) -> bool:
    """Function to delete an object from s3 based on its key

    :param key: s3 key of object to delete
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :return: True if delete successful, else return False
    """
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    key_exists_before = key_exists_s3(key, s3_client=s3_client, s3_params=s3_params)
    try:
        # delete object
        response = s3_client.delete_object(Bucket=s3_params.bucket_name, Key=key)
        key_exists_after = key_exists_s3(key, s3_client=s3_client, s3_params=s3_params)
        # if key exists before but not after, we know it has been deleted and we can return True
        if key_exists_before and not key_exists_after: return True
    except Exception as e:
        logger.debug(f"Failed to delete S3 object: {key}")
    
    return False


def delete_s3_objects_by_prefix(
    prefix: str, 
    s3_client = None, 
    s3_params: S3Params = None,
    **kwargs
) -> List[str]:
    """Function to delete all objects in S3 bucket that share a given prefix

    :param prefix: S3 Bucket Path Prefix
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :return: List of S3 Object Keys that were successfully delete
    """   
    # Adding trailing slash if looking for "directory"
    if prefix.strip() != "" and not prefix.endswith("/"):
        old_prefix = prefix
        prefix = f"{prefix}/"
        logger.debug(f"Old Prefix {old_prefix} -> New Prefix: {prefix}")

    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    deleted_keys: List[str]= []

    try:
        # get list of valid keys using input prefix
        valid_keys = get_valid_s3_keys(prefix, s3_client=s3_client, s3_params=s3_params)
        num_keys = len(valid_keys)

        # iteratively delete keys
        for i, key in enumerate(valid_keys):
            is_deleted = delete_s3_object(key, s3_client=s3_client, s3_params=s3_params)
            # if successfully deleted print message and add to deleted_keys list
            if is_deleted:
                logger.debug(f"Deleted S3 file: {i+1}/{num_keys} -- {key}")
                deleted_keys.append(key)
    except Exception as e:
        logger.debug(f"Failed to delete S3 object by prefix: {prefix}")
        
    return deleted_keys


def extract_compressed_file_s3(
    key: str, 
    extracted_dir_prefix: Optional[str] = None,
    s3_client = None, 
    s3_params: S3Params =  None,
    **kwargs
) -> str:
    """Function to extract zip file contents in S3

    Compression Support:
    - .zip
    - .gzip [TODO]
    - .tar.gz [TODO]

    :param key: key to compressed file
    :param extracted_dir_prefix: directory to prefix the extracted archive, defaults to None
    :param s3_client: provided s3 client object, defaults to None
    :param s3_params: provided instance of S3Params, defaults to None
    :return: prefix path to directory where uncompressed file contents exist
    """
    # use default if not provided
    if s3_params is None: s3_params = get_s3_default_params()
    # create session client if not provided
    if s3_client is None: s3_client = get_s3_session_client(s3_params)

    def _decompress_zip_and_upload_to_s3(buffer: BytesIO, dir_path: str, zip_fpath: str = None) -> List[str]:
        """Inner function to unzip .zip file in S3 and upload unzipped files
        back to S3

        :param buffer: zip file buffer
        :param dir_path: path to parent directory of key .zip file
        :param zip_fpath: path to zip file, defaults to None 
        :return: list of uploaded S3 URI keys
        """
        ret = []
        try:
            # new ZipFile instance
            with zipfile.ZipFile(buffer) as z:
                for i, filename in enumerate(z.namelist()): 
                    # ignore directories and MACOSX specific things
                    if filename.endswith("/") or "MACOSX" in filename: continue
                    logger.debug(f"Extracting File Name: {filename}")
                    # get zipfile info
                    file_info = z.getinfo(filename)
                    # create new save key
                    save_key = os.path.join(dir_path, filename)   
                    # Decompress and copy the files to the 'unzipped' S3 folder 
                    uploaded_key = upload_to_s3(
                        z.open(file_info).read(), 
                        save_key, 
                        s3_client=s3_client, 
                        s3_params=s3_params
                    )
                    # remove s3://<bucket> from upload key
                    if "s3://" in uploaded_key:
                        parsed_uri = urlparse(uploaded_key)
                        uploaded_key = parsed_uri.path
                        if uploaded_key.startswith("/"):
                            uploaded_key = uploaded_key[1:]

                    # if nested .zip files, recursively extract
                    if filename.endswith(".zip"):
                        # parse new dir path
                        new_dir_path = uploaded_key.split("/")[:-1]
                        new_dir_path = os.path.join(*new_dir_path)
                        if dir_path.startswith("/") and not new_dir_path.startswith("/"):
                            new_dir_path = f"/{new_dir_path}"
                        # add trailing slash if does not exist
                        if not new_dir_path.endswith("/"):
                            new_dir_path = f"{new_dir_path}/"
                        
                        logger.debug(f"Nested .zip File Found. New Dir Path: {new_dir_path}")
                        buffer = get_s3_file(uploaded_key, s3_client=s3_client, bytes_io=True)
                        new_path_li = _decompress_zip_and_upload_to_s3(buffer, new_dir_path, zip_fpath=uploaded_key)
                        ret.extend(new_path_li)
                    else:
                        ret.append(uploaded_key)
        except Exception as e:
            logger.error(f"ERROR! ***DecompressZipFileError*** for file: {zip_fpath}. {e}")
        
        return ret

    # read in compressed file
    buffer = get_s3_file(key, s3_client=s3_client, s3_params=s3_params, bytes_io=True)

    # get last directory path
    path_parts = key.strip("/").split("/")
    dir_path = ""
    fname = path_parts[-1]
    if len(path_parts) > 1: 
        dir_path = os.path.join(*path_parts[:-1])
    
    # add extracted dir prefix if passed in
    if extracted_dir_prefix:
        dir_path = os.path.join(dir_path, extracted_dir_prefix)

    # get original file name index in split path
    original_file_index = path_parts.index(fname)
    path_offset = 1
    if extracted_dir_prefix:
        path_offset = 2
    
    # get original file name index in split path
    original_file_index = path_parts.index(fname)
    
    # handle different compressions
    new_keys = []
    if key.endswith(".zip"):
        new_keys = _decompress_zip_and_upload_to_s3(buffer, dir_path, zip_fpath=key)
    elif key.endswith(".gzip"):
        pass
    elif key.endswith(".tar.gz"):
        pass

    logger.debug(f"New Extracted Keys: {new_keys[:10]}, ... +{len(new_keys)-10}")

    # get archive dir
    if len(new_keys) < 1:
        logger.warning(f"WARNING: extract_compressed_file_s3() resulted in 0 Extracted Keys!")
        return
    
    # get last directory path
    archive_dirpath = None
    for k in new_keys:
        # split key by path
        new_path_parts = k.strip("/").split("/")
        # we need split path to have at least the original compressed file level
        if len(new_path_parts) < original_file_index+path_offset: continue
        # format the resulting archive dirpath to return
        archive_dirpath = os.path.join(*new_path_parts[:original_file_index+path_offset])
        # make sure archive_dirpath ends with "/"
        # NOTE: this is necessary for S3 prefixes
        if not archive_dirpath.endswith("/"): archive_dirpath =f"{archive_dirpath}/"
        # break if found. We only need one.
        break
    
    # return the archive dir/prefix path
    return archive_dirpath
